use nom::{
    bytes::complete::{tag, take_until, take_until1, take_while, take_while1},
    character::complete::char,
    multi::many0,
    sequence::{preceded, separated_pair},
    Parser,
};
use std::collections::HashMap;

pub struct TemplateWorkspaceEnvironment {
    pub id: usize,

    /// Workspace environment name.
    pub name: String,

    /// Variables in this environment.
    pub variables: HashMap<String, TemplateEnvironmentVariable>,
}

pub struct TemplateEnvironmentVariable {
    pub id: usize,
    pub name: String,
    pub default_value: String,
    pub current_value: Option<String>,
    pub secret: bool,
}

pub struct TemplateWorkspace {
    pub id: usize,
    /// Workspace environment variables accessible by all
    /// child entries.
    pub environments: Vec<TemplateWorkspaceEnvironment>,

    pub entries: Vec<TemplateEntry>,
}

pub enum TemplateEntry {
    Directory(TemplateDirectory),
    Request(TemplateRequest),
}

pub struct TemplateDirectory {
    pub id: usize,
    /// Directory variables accessible only to child entries
    /// of this directory.
    pub variables: HashMap<String, TemplateEnvironmentVariable>,

    pub entries: Vec<TemplateEntry>,
}

pub struct TemplateRequest {
    pub id: usize,

    pub method: String,

    /// Template display name
    pub name: String,

    /// The request URL
    pub url: String,

    /// Path parameters used to substitute path segments.
    pub path_params: HashMap<String, String>,

    pub headers: Vec<(String, String)>,
}

/// A fully deconstructed URL from a template request.
/// Used as an intermediate step for populating the final URL with variables.
#[derive(Debug)]
pub struct TemplateRequestUrl<'a> {
    /// The URL scheme, e.g. `http`.
    pub scheme: &'a str,

    /// The URL host, includes the port if specified.
    pub host: &'a str,

    /// The URL path segments.
    ///
    /// All segments will be formatted as `/segment`, meaning empty Static
    /// fields represent a `/`, which is usually trailing.
    pub path: Vec<Segment<'a>>,

    /// Query parameters.
    pub query_params: Vec<(&'a str, &'a str)>,
}

impl<'a> TemplateRequestUrl<'a> {
    pub fn parse(input: &'a str) -> Result<Self, nom::Err<nom::error::Error<&'a str>>> {
        let (input, scheme) = take_while1(char::is_alphabetic)(input)?;

        let (input, _) = tag("://")(input)?;

        let mut path_parser = many0(preceded(
            char('/'),
            take_while(|c: char| c.is_ascii_alphanumeric() || c == ':'),
        ));

        let result = take_until1::<_, _, nom::error::Error<_>>("?")(input);
        match result {
            // URL has query parameters
            Ok((query, path)) => {
                // Parse query
                // First char will always be a '?' since we parsed succesfully
                let mut query = &query[1..];
                let mut query_params = vec![];

                loop {
                    if query.is_empty() {
                        break;
                    }

                    let (i, params) = separated_pair(
                        take_while(|c: char| c != '='),
                        char('='),
                        take_while(|c: char| c != '&'),
                    )
                    .parse(query)?;

                    query = i;
                    query_params.push((params.0, params.1));

                    if let Ok((i, _)) = char::<_, nom::error::Error<_>>('&').parse(query) {
                        query = i;
                    }
                }

                debug_assert!(query.is_empty());

                // Check path segments

                match take_until::<_, _, nom::error::Error<_>>("/")(path) {
                    // Path exists
                    Ok((path, host)) => {
                        let (input, segments) = path_parser.parse(path)?;
                        debug_assert!(input.is_empty());
                        Ok(TemplateRequestUrl {
                            scheme,
                            host,
                            path: segments
                                .into_iter()
                                .map(|segment| {
                                    segment
                                        .strip_prefix(':')
                                        .map_or(Segment::Static(segment), Segment::Dynamic)
                                })
                                .collect(),
                            query_params,
                        })
                    }

                    // No path segments
                    Err(_) => Ok(TemplateRequestUrl {
                        scheme,
                        host: path,
                        path: vec![],
                        query_params,
                    }),
                }
            }
            // No query params
            Err(_) => {
                match take_until::<_, _, nom::error::Error<_>>("/")(input) {
                    // Path exists
                    Ok((path, host)) => {
                        let (input, segments) = path_parser.parse(path)?;
                        debug_assert!(input.is_empty());
                        Ok(TemplateRequestUrl {
                            scheme,
                            host,
                            path: segments
                                .into_iter()
                                .map(|segment| {
                                    segment
                                        .strip_prefix(':')
                                        .map_or(Segment::Static(segment), Segment::Dynamic)
                                })
                                .collect(),
                            query_params: vec![],
                        })
                    }
                    // No path segments
                    Err(_) => Ok(TemplateRequestUrl {
                        scheme,
                        host: input,
                        path: vec![],
                        query_params: vec![],
                    }),
                }
            }
        }
    }
}

#[derive(Debug, PartialEq, Eq)]
pub enum Segment<'a> {
    /// Path segments that do not change.
    /// The value is the final path value.
    Static(&'a str),

    /// Path segments that depend on request configuration.
    /// The value is the name of the variable in the request configuration
    /// that contains the final path value.
    Dynamic(&'a str),
}

#[cfg(test)]
mod tests {
    use super::{Segment, TemplateRequestUrl};

    #[test]
    fn parses_path_placeholders() {
        let input = "http://localhost:4000/foo/:bar/bax";

        let expected_path = vec![
            Segment::Static("foo"),
            Segment::Dynamic("bar"),
            Segment::Static("bax"),
        ];

        let url = TemplateRequestUrl::parse(input).unwrap();

        assert_eq!("http", url.scheme);
        assert_eq!("localhost:4000", url.host);
        assert_eq!(expected_path, url.path);
        assert!(url.query_params.is_empty());
    }

    #[test]
    fn parses_path_placeholders_trailing_slash() {
        let input = "http://localhost:4000/foo/:bar/bax/";

        let expected_path = vec![
            Segment::Static("foo"),
            Segment::Dynamic("bar"),
            Segment::Static("bax"),
            Segment::Static(""),
        ];

        let url = TemplateRequestUrl::parse(input).unwrap();

        assert_eq!("http", url.scheme);
        assert_eq!("localhost:4000", url.host);
        assert_eq!(expected_path, url.path);
        assert!(url.query_params.is_empty());
    }

    #[test]
    fn parses_no_path_segments() {
        let input = "http://localhost:4000";

        let url = TemplateRequestUrl::parse(input).unwrap();

        assert_eq!("http", url.scheme);
        assert_eq!("localhost:4000", url.host);
        assert!(url.path.is_empty());
        assert!(url.query_params.is_empty());
    }

    #[test]
    fn parse_no_path_segments_trailing_slash() {
        let input = "http://localhost:4000/";

        let url = TemplateRequestUrl::parse(input).unwrap();

        assert_eq!("http", url.scheme);
        assert_eq!("localhost:4000", url.host);
        assert_eq!(vec![Segment::Static("")], url.path);
        assert!(url.query_params.is_empty());
    }

    #[test]
    fn parse_query_params_no_path() {
        let input = "http://localhost:4000?foo=bar&baz=bax";

        let url = TemplateRequestUrl::parse(input).unwrap();

        assert_eq!("http", url.scheme);
        assert_eq!("localhost:4000", url.host);
        assert!(url.path.is_empty());
        assert_eq!(vec![("foo", "bar"), ("baz", "bax")], url.query_params);
    }

    #[test]
    fn parse_query_params_with_path() {
        let input = "http://localhost:4000/foo/:bar?foo=bar&baz=bax";

        let url = TemplateRequestUrl::parse(input).unwrap();

        assert_eq!("http", url.scheme);
        assert_eq!("localhost:4000", url.host);
        assert_eq!(
            vec![Segment::Static("foo"), Segment::Dynamic("bar")],
            url.path
        );
        assert_eq!(vec![("foo", "bar"), ("baz", "bax")], url.query_params);
    }
}

